import sys
import getopt
import math
import numpy
import time
import matplotlib.pyplot as plt
import matplotlib.animation as animate

# TODO:	    Colormap should respond to changes in index array.

def unpack_line(r, np, nq):
    
    tt = float(r[0])
    ii = []
    mm = []
    xx = []
    yy = []
    zz = []
    for j in range(np):
        ii.append(int(r[nq*j+1]))
        if nq == 4:
            mm.append(2.**(-j))
        else:
            mm.append(float(r[nq*j+2]))
        xx.append(float(r[nq*j+nq-2]))
        yy.append(float(r[nq*j+nq-1]))
        zz.append(float(r[nq*j+nq]))

    return tt, ii, mm, xx, yy, zz

def interpolate(xp, xx, tfac):
    
    xint = []
    for j in range(len(xx)):
        xint.append(xp[j] + tfac*(xx[j]-xp[j]))
    return xint

def read_data(filename, dt, N, have_mass):

    # Read in the data file.  Format is defined in smalln.

    print('reading data from', filename)
    sys.stdout.flush()

    f = open(filename)
    ll = f.readlines()
    f.close()

    nt = len(ll)
    print(nt, 'records read')
    sys.stdout.flush()

    # Do some preliminary analysis.

    r = ll[0].split()			# trust the first line
    if have_mass:
        nq = 5
    else:
        nq = 4
    np = (len(r)-1)/nq
    print('np =', np, 'nq =', nq)
    
    tt, ii, mm, xx, yy, zz = unpack_line(r, np, nq)
    tnext = tt

    if N > 0:				# N trumps dt
        r = ll[nt-1].split()		# last line
        tlast, dum, dum, dum, dum = unpack_line(r, np, nq)
        pdt = False
        if dt > 0: pdt = True
        dt = (tlast-tt)/N
        if pdt: print('resetting dt =', dt)

    t = []
    i = []
    m = []
    x = []
    y = []
    z = []
    line = 0
    short = 0

    for l in ll:

        r = l.split()
        line += 1

	# Format is: time  id1 m1 x1 y1 z1  id2 m2 x2 y2 z2  id3 m3 x3 y3 z3 ...

        if nq*np + 1 == len(r):

            # Save old data and unpack the line.

            tp = tt
            mp= mm
            xp = list(xx)
            yp = list(yy)
            zp = list(zz)
            tt, ii, mm, xx, yy, zz = unpack_line(r, np, nq)

            while tt >= tnext:

                if line == 1 or dt == 0.0:

                    tint = tt
                    mint = mm
                    xint = list(xx)
                    yint = list(yy)
                    zint = list(zz)
                    
                else:
                    
                    # Interpolate to time tnext.

                    tint = tnext
                    tfac = (tnext-tp)/(tt-tp)
                    mint = interpolate(mp, mm, tfac)
                    xint = interpolate(xp, xx, tfac)
                    yint = interpolate(yp, yy, tfac)
                    zint = interpolate(zp, zz, tfac)

                # Append the data to the master arrays.

                t.append(tint)
                i.append(ii)
                m.append(mint)
                x.append(xint)
                y.append(yint)
                z.append(zint)

                tnext += dt
                if dt == 0.0: break

        else:				# temporary 2-body treatment
            
            #print 'ignoring short record:', len(r), '!=', 4*np+1
            short += 1

    ta = numpy.array(t)
    ia = numpy.array(i)
    ma = numpy.array(m)
    xa = numpy.array(x)			# should be nt x np
    ya = numpy.array(y)
    za = numpy.array(z)

    nt = len(ta)
    print('nt =', nt)
    print(short, 'short records')
    #print 'xa.shape =', xa.shape	# should be nt x np
    return nt, ta, ia, ma, xa, ya, za

def print_help():
    print('keyboard controls:')
    print('    space    pause/resume')
    print('    a        expand view to contain all particl;es')
    print('    h        print this message')
    print('    q        quit')
    print('    z        zoom in')
    print('    Z        zoom out')
    print('    right    pan right')
    print('    left     pan left')
    print('    up       pan up')
    print('    down     pan down')
    print('    <        first frame')
    print('    >        last frame')
    print('mouse click reverses direction')

# We seem to need some of these global for the animate functions to work...

nf = 0
df = 1
xmin = 0.0
xmax = 0.0
ymin = 0.0
ymax = 0.0
anim_running = True
current_frame = 0
shift = 0.25
zoom = 1.5

def animate_data(t, m, x, y, id, lx, ly, scale, delay):

    global xmin, xmax, ymin, ymax

    print('animating data')
    colormap = ['r', 'y', 'b', 'm', 'g', 'c', 'k', 'k', 'k', 'k', 'k', 'k']

    # Determine length scales.

    xmin = numpy.min(x[0,:])
    xmax = numpy.max(x[0,:])
    dx = xmax - xmin
    xav = 0.5*(xmin+xmax)
    ymin = numpy.min(y[0,:])
    ymax = numpy.max(y[0,:])
    dy = ymax - ymin
    yav = 0.5*(ymin+ymax)

    # Allow modification of overall scale by user-specified scale factor.

    dx = 5*max(dx, dy)*scale
    xmin = xav - 0.5*dx
    xmax = xav + 0.5*dx
    ymin = yav - 0.5*dx
    ymax = yav + 0.5*dx

    # Determine time scales.

    dtscale = t[-1]
    if t[0] < 0: dtscale = max(dtscale, -t[0])
    nf = int(math.floor(math.log10(dtscale)))
    #print 'dtscale =', dtscale, 'nf =', nf
    if nf > 3:
        tformat = '%'+str(nf+2)+'.0f'
    elif nf > 0:
        tformat = '%'+str(nf+4)+'.2f'
    else:
        tformat = '%'+str(-nf+7)+'.'+str(-nf+4)+'f'
    lformat = tformat+'/'+tformat+'    frame %d/%d'
    s = numpy.log(list(m[0,:]))
    smin = numpy.min(s)
    smax = numpy.max(s)
    s = 5 + 25*(s-smin)/(smax-smin)    # sizes logarithmic in mass, range 5-30
    print('s =', s)
    nt,np = x.shape
    fig = plt.figure()
    
    scat = plt.scatter(x[0,:], y[0,:], c=colormap[:np], s=s)
    #plt.axis('square')	# fails in ubuntu
    plt.xlabel(lx)
    plt.ylabel(ly)
    plt.xlim(xmin, xmax)
    plt.ylim(ymin, ymax)
    plt.grid()

    def update(frame):

        # What idiot decided that the syntax for set_offsets should
        # be so different from the syntax for scatter??

        off = []
        for j in range(np):
            off.append(x[frame,j])
            off.append(y[frame,j])
        scat.set_offsets(off)
        text = fig.suptitle('')
        text.set_text('')
        text.set_text('time '+lformat%(t[frame],t[-1], frame, nt-1))

        return scat,

    def nextframe(nt):			# step nf forward or backward by df
        global nf, df			# stay fixed at nf = 0 or nt-1
        global current_frame
        nf = 0
        while nf < nt:
            current_frame = nf
            yield nf
            nf += df
            if nf < 0: nf = 0
            if nf >= nt: nf = nt-1
    
    def new_limits(x, y, fac):
        xmin = x.min()
        xmax = x.max()
        ymin = y.min()
        ymax = y.max()
        xm = 0.5*(xmin+xmax)
        dx = xmax - xm
        xmax = xm + fac*dx
        xmin = xm - fac*dx
        ym = 0.5*(ymin+ymax)
        dy = ymax - ym
        ymax = ym + fac*dy
        ymin = ym - fac*dy
        return xmin, xmax, ymin, ymax

    def onClick(event):			# reverse direction on mouse click
        global nf, df
        df = -df

    def onKey(event):			# manage key presses
        global anim_running
        global nf, df
        global xmin, xmax, ymin, ymax
        global current_frame
        global shift, zoom
        if event.key == 'a':		# a = reset limits to current particles
            xmin, xmax, ymin, ymax = \
			new_limits(x[current_frame],
                                   y[current_frame], 1.25)
            plt.xlim(xmin, xmax)
            plt.ylim(ymin, ymax)
        elif event.key == 'q':		# q = quit
            sys.exit(0)
        elif event.key == 'z':		# z = zoom in
            xm = 0.5*(xmin+xmax)
            dx = xmax - xm
            xmax = xm + dx/zoom
            xmin = xm - dx/zoom
            ym = 0.5*(ymin+ymax)
            dy = ymax - ym
            ymax = ym + dy/zoom
            ymin = ym - dy/zoom
            plt.xlim(xmin, xmax)
            plt.ylim(ymin, ymax)
        elif event.key == 'Z':		# Z = zoom out
            xm = 0.5*(xmin+xmax)
            dx = xmax - xm
            xmax = xm + dx*zoom
            xmin = xm - dx*zoom
            ym = 0.5*(ymin+ymax)
            dy = ymax - ym
            ymax = ym + dy*zoom
            ymin = ym - dy*zoom
            plt.xlim(xmin, xmax)
            plt.ylim(ymin, ymax)
        elif event.key == ' ':		# space = pause/restart
            if anim_running:
                anim.event_source.stop()
                anim_running = False
            else:
                anim.event_source.start()
                anim_running = True
        elif event.key == 'right':
            dx = shift*(xmax - xmin)
            xmax += dx
            xmin += dx
            plt.xlim(xmin, xmax)
        elif event.key == 'left':
            dx = shift*(xmax - xmin)
            xmax -= dx
            xmin -= dx
            plt.xlim(xmin, xmax)
        elif event.key == 'up':
            shift = 0.5
            dy = shift*(ymax - ymin)
            ymax += dy
            ymin += dy
            plt.ylim(ymin, ymax)
        elif event.key == 'down':
            shift = 0.5
            dy = shift*(ymax - ymin)
            ymax -= dy
            ymin -= dy
            plt.ylim(ymin, ymax)
        elif event.key == '<':
            nf = 0
            df = 1
        elif event.key == '>':
            nf = nt - 1
            df = -1
        elif event.key == 'h':
            print_help()
        else:
            print('key =', event.key)

    fig.canvas.mpl_connect('key_press_event', onKey)
    fig.canvas.mpl_connect('button_press_event', onClick)
    anim = animate.FuncAnimation(fig, update, frames=nextframe(nt),
                                 interval=delay, repeat=False)
    plt.show()

def plot_data(t, x, lx, ly):
    nt,np = x.shape
    fig = plt.figure()
    plt.plot(t, x, 'b-')
    plt.xlabel(lx)
    plt.ylabel(ly)
    plt.grid()
    plt.show()

def write_data(t, x, y, i, filename):
    nt,np = x.shape
    f = open(filename)
    for i in range(nt):
        s = str(t[i])+' '
        for j in range(np):
            s += str(i(i,j))+' '+str(x(i,j))+' '+str(y(i,j))+' 0.0\n'
        f.write(s)
    f.close()

if __name__ == '__main__':

    anim = True
    dt = 0.0
    delay = 30
    file = 'abc.dat'
    have_mass = True
    N = 0
    outfile = None
    proj = 3
    scale = 1.5

    try:
        opts, args = getopt.getopt(sys.argv[1:], "ad:D:f:mN:o:p:s:")
    except getopt.GetoptError as err:
        print(str(err))
        sys.exit(1)

    for o, a in opts:
        if o == "-a":
            anim = not anim
        elif o == "-d":
            dt = float(a)
        elif o == "-D":
            delay = int(a)
        elif o == "-f":
            file = a
        elif o == "-m":
            have_mass = not have_mass
        elif o == '-N':
            N = int(a)
        elif o == "-o":
            outfile = a
        elif o == '-p':
            proj = int(a)
        elif o == "-s":
            scale = float(a)
        else:
            print("unexpected argument", o)
            sys.exit(1)

    nt, t, i, m, x, y, z = read_data(file, dt, N, have_mass)

    # Animate or plot the data.

    if anim:
        a1 = x
        a2 = y
        l1 = 'x'
        l2 = 'y'
        if proj == 1:
            a1 = y
            a2 = z
            l1 = 'y'
            l2 = 'z'
        elif proj == 2:
            a2 = z
            l2 = 'z'
        animate_data(t, m, a1, a2, i[:nt], l1, l2, scale, delay)
    else:
        plot_data(t, m, x, 't', 'x')



from pylab import *
import sys
import re

if __name__ == '__main__':
    if len(sys.argv) != 3:
        print("usage: python binary_hist.py <input filename> <time of histogram>")
        sys.exit(1)
    else:
        fname = sys.argv[1]
        time = float(sys.argv[2])
        f = open(fname, "r")
        inblock = False
        EkTs = []
        for line in f:
            if re.search("%%% time= (\d+\.\d*)", line):
                if float(re.search("%%% time= (\d+\.\d*)", line).group(1)) == time:
                    inblock = True
            if inblock and re.search("%%% .*E/kT=(-?\d+\.\d+)", line):
                EkTs.append(float(re.search("%%%.*E/kT=(-?\d+\.\d+)",
                                            line).group(1)))
    
            if inblock and re.search("%%% Emul/E", line):
                inblock = False
        f.close()
    
        if len(EkTs) > 0:
            hist(EkTs)
            show()
        else:
            print("No binaries found at time = %f." % time)
    



from amuse.community import *
from amuse.community.interface.gd import GravitationalDynamics
from amuse.community.interface.gd import GravitationalDynamicsInterface
from amuse.community.interface.gd import GravityFieldInterface
from amuse.community.interface.gd import GravityFieldCode

# *** This script, together with the defaults in
# *** GravitationalDynamicsInterface, will be used to generate both
# *** the header file interface.h and the stub interface.cc.  Since
# *** interface.cc has been  hand-coded to implement the details,
# *** MAKE SURE TO SAVE IT SOMEWHERE, as build.py can overwrite it!

class ph4Interface(CodeInterface,
                   LiteratureReferencesMixIn,
                   GravitationalDynamicsInterface,
                   StoppingConditionInterface,
                   GravityFieldInterface):
    """
    Parallel, GPU-accelerated, N-body integration module with block
    time steps, using a 4th-order Hermite integration scheme.

        .. [#] [2018araa.book.....P] Portegies Zwart, S. & McMillan, S.L.W., 2018    
    """

    # Interface specification.

    include_headers = ['interface.h', 'stopcond.h']

    MODE_GPU = 'gpu'
    MODE_CPU = 'cpu'
    
    def __init__(self, mode = MODE_CPU, **options):
        CodeInterface.__init__(
            self,
            name_of_the_worker=self.name_of_the_muse_worker(mode),
            **options
        )
        LiteratureReferencesMixIn.__init__(self)

    # Interface functions:
    @legacy_function
    def new_particle():
        """
        Define a new particle in the stellar dynamics code. The
        particle is initialized with the provided mass, radius,
        position and velocity. This function returns an index that
        can be used to refer to this particle.
        """
        function = LegacyFunctionSpecification()
        function.can_handle_array = True
        function.addParameter('index_of_the_particle', dtype='int32',
                              direction=function.OUT, description =
            """
            An index assigned to the newly created particle.
            This index is supposed to be a local index for the code
            (and not valid in other instances of the code or in other codes)
            """,
            unit = INDEX
            )

        function.addParameter('mass', dtype='float64', direction=function.IN,
                 description = "The mass of the particle", 
                 unit = nbody_system.mass)
        function.addParameter('x', dtype='float64', direction=function.IN,
                 description = "The initial position vector of the particle",
                 unit = nbody_system.length)
        function.addParameter('y', dtype='float64', direction=function.IN,
                 description = "The initial position vector of the particle",
                 unit = nbody_system.length)
        function.addParameter('z', dtype='float64', direction=function.IN,
                 description = "The initial position vector of the particle",
                 unit = nbody_system.length)
        function.addParameter('vx', dtype='float64', direction=function.IN,
                 description = "The initial velocity vector of the particle",
                 unit = nbody_system.speed)
        function.addParameter('vy', dtype='float64', direction=function.IN,
                 description = "The initial velocity vector of the particle",
                 unit = nbody_system.speed)
        function.addParameter('vz', dtype='float64', direction=function.IN,
                 description = "The initial velocity vector of the particle",
                 unit = nbody_system.speed)
        function.addParameter('radius', dtype='float64', direction=function.IN,
                 description = "The radius of the particle",
                 default = -1, unit = nbody_system.length)
        function.addParameter('id', dtype='int32', direction=function.IN,
                 description = "Identifier of the particle, "
                               +"option for restoring state after loading",
                 default = -1, unit = NO_UNIT)
        function.result_type = 'int32'
        function.result_doc = """ 0 - OK
            particle was created and added to the model
        -1 - ERROR
            particle could not be created"""
        return function
    
    @legacy_function
    def get_particle_timestep():
        """
        Retrieve the timestep of a particle.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('index_of_the_particle',
                              dtype='int32',
                              direction=function.IN,
                              description = "Index of the particle to get the timestep from. This index must have been returned by an earlier call to :meth:`new_particle`")
        function.addParameter('timestep',
                              dtype='float64',
                              unit=nbody_system.time,
                              direction=function.OUT,
                              description = "The current timestep of the particle")
        function.result_type = 'int32'
        function.can_handle_array = True
        return function
    
    def name_of_the_muse_worker(self, mode):
        if mode == self.MODE_CPU:
            return 'ph4_worker'
        elif mode == self.MODE_GPU:
            return 'ph4_worker_gpu'
        else:
            return 'ph4_worker'
        
    # Inheritance from GravitationalDynamicsInterface means that
    # functions in the standard interface don't need to be defined.
    # See interface.py.2 for a laboriously hand-coded version written
    # before I discovered this fact!    (Steve McMillan, 10/10)

    # Additional functions defined here will be reflected in
    # interface.h and must be provided in interface.cc in order for
    # ph4_worker to build.

    # The following functions aren't defined in the default interface:
    @legacy_function
    def set_time():
        """
        Set the model time. Should use set_begin_time!
        """
        function = LegacyFunctionSpecification()
        function.addParameter('time', dtype='float64', direction=function.IN,
            description = "The model time to start at", unit = nbody_system.time)
        function.result_type = 'int32'
        function.result_doc = """
        0 - OK
            Time value was changed
        -2 - ERROR
            The code does not support setting the  time
        """
        return function
        
    @legacy_function
    def set_eta():
        """
        Set the current accuracy parameter.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('eta', dtype='float64',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_eta():
        """
        Get the current accuracy parameter.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('eta', dtype='float64',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_gpu():
        """
        Set use_gpu.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('gpu', dtype='int32',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_gpu():
        """
        Get use_gpu.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('gpu', dtype='int32',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_gpu_id():
        """
        Set gpu_id.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('gpu_id', dtype='int32',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_gpu_id():
        """
        Get gpu_id.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('gpu_id', dtype='int32',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_manage_encounters():
        """
        Set the value of manage_encounters.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('manage_encounters', dtype='int32',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_manage_encounters():
        """
        Get the value of manage_encounters.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('manage_encounters', dtype='int32',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_zero_step_mode():
        """
        Set the value of zero_step_mode.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('zero_step_mode', dtype='int32',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_zero_step_mode():
        """
        Get the value of zero_step_mode.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('zero_step_mode', dtype='int32',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_force_sync():
        """
        Set the value of force_sync.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('force_sync', dtype='int32',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_force_sync():
        """
        Get the value of force_sync.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('force_sync', dtype='int32',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_sync_time():
        """
        Set the value of sync_time.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('sync_time', dtype='float64',
                              direction=function.IN, unit = nbody_system.time)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_sync_time():
        """
        Get the value of sync_time.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('sync_time', dtype='float64',
                              direction=function.OUT, unit = nbody_system.time)
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_block_steps():
        """
        Set the value of block_steps.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('set_block_steps', dtype='int32',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_block_steps():
        """
        Get the value of block_steps.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('get_block_steps', dtype='int32',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_total_steps():
        """
        Set the value of total_steps.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('set_total_steps', dtype='int32',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_total_steps():
        """
        Get the value of total_steps.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('get_total_steps', dtype='int32',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_number_of_particles_updated():
        """
        Return the number of particles added or deleted during the last evolve.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function
    
    @legacy_function
    def get_id_of_updated_particle():
        """
        Return the number of particles added or deleted during the last evolve.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('index_of_update', dtype='int32',
                              direction=function.IN, 
                 description = 'index in the updated particles list')
        function.addParameter('index_of_particle', dtype='int32',
                              direction=function.OUT)
        function.addParameter('kind_of_update', dtype='int32',
                              direction=function.OUT,
                 description = 'kind of update (2, addition), (1, deletion)')
        function.can_handle_array = True
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_binary_energy():
        """
        Return the total energy in all binaries.
        """
        function = LegacyFunctionSpecification()
        function.addParameter(
            'binary_energy', 
            dtype='float64',
            direction=function.OUT,
            unit = nbody_system.energy
        )
        function.result_type = 'int32'
        return function

    @legacy_function
    def set_initial_timestep_fac():
        """
        Set the current accuracy parameter.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('initial_timestep_fac', dtype='float64',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_initial_timestep_fac():
        """
        Get the current accuracy parameter.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('initial_timestep_fac', dtype='float64',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function
    
    @legacy_function
    def set_initial_timestep_limit():
        """
        Set the current accuracy parameter.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('initial_timestep_limit', dtype='float64',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_initial_timestep_limit():
        """
        Get the current accuracy parameter.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('initial_timestep_limit', dtype='float64',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function
    
    @legacy_function
    def set_initial_timestep_median():
        """
        Set the current accuracy parameter.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('initial_timestep_median', dtype='float64',
                              direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_initial_timestep_median():
        """
        Get the current accuracy parameter.
        """
        function = LegacyFunctionSpecification()
        function.addParameter('initial_timestep_median', dtype='float64',
                              direction=function.OUT)
        function.result_type = 'int32'
        return function
    
    @legacy_function
    def recompute_timesteps():
        """
        Force recomputation of all timesteps (assume a synchronized system),
        """
        function = LegacyFunctionSpecification()
        function.result_type = 'int32'
        return function


class ph4(GravitationalDynamics,GravityFieldCode):

    # The actual module.

    def __init__(self, convert_nbody = None, **keyword_arguments):
        legacy_interface = ph4Interface(**keyword_arguments)

        self.stopping_conditions = StoppingConditions(self)
        
        GravitationalDynamics.__init__(self,
                                       legacy_interface,
                                       convert_nbody,
                                       **keyword_arguments)
                                       
    def define_state(self, handler):
        GravitationalDynamics.define_state(self, handler)
        handler.add_method('RUN', 'get_particle_timestep')
        GravityFieldCode.define_state(self, handler)
        
        handler.add_method('EDIT', 'set_state')
        handler.add_method('EDIT', 'set_velocity')
        handler.add_method('EDIT', 'set_mass')
        handler.add_method('EDIT', 'set_position')
        handler.add_method('CHANGED','before_get_parameter')
        
        handler.add_transition('RUN', 'CHANGED', 'set_state', False)
        handler.add_transition('RUN', 'CHANGED', 'set_velocity', False)
        handler.add_transition('RUN', 'CHANGED', 'set_mass', False)
        handler.add_transition('RUN', 'CHANGED', 'set_position', False)
        handler.add_transition('CHANGED', 'RUN', 'synchronize_model')
        handler.add_method('CHANGED', 'get_state')
        handler.add_method('CHANGED', 'get_mass')
        handler.add_method('CHANGED', 'get_position')
        handler.add_method('CHANGED', 'get_velocity')
        handler.add_method('CHANGED', 'get_particle_timestep')
        
        self.stopping_conditions.define_state(handler)

    def define_parameters(self, handler):

        # Set/get parameters specific to the module, not part of the
        # standard interface.  Accessors used here must be defined
        # above and reflected in interface.cc.  Python access is
        # (e.g.)
        #
        #        ph4.parameters.timestep_parameter = xxx

        handler.add_method_parameter(
            "get_eta",                   # getter name in interface.cc
            "set_eta",                   # setter name in interface.cc
            "timestep_parameter",        # python parameter name
            "timestep parameter",        # description
            default_value = 0.14
        )

        handler.add_method_parameter(
            "get_eps2",                  # already defined in standard interface
            "set_eps2",                  # already defined in standard interface
            "epsilon_squared", 
            "smoothing parameter for gravity calculations", 
            default_value = 0.0 | nbody_system.length * nbody_system.length
        )

        handler.add_method_parameter(
            "get_gpu",                   # getter name in interface.cc
            "set_gpu",                   # setter name in interface.cc
            "use_gpu",                   # python parameter name
            "use GPU",                   # description
            default_value = 1
        )
        
        handler.add_method_parameter(
            "get_gpu_id",                # getter name in interface.cc
            "set_gpu_id",                # setter name in interface.cc
            "gpu_id",                    # python parameter name
            "GPU ID",                    # description
            default_value = -1
        )
        
        handler.add_method_parameter(
            "get_manage_encounters",     # getter name in interface.cc
            "set_manage_encounters",     # setter name in interface.cc
            "manage_encounters",	 # python parameter name
            "manage close encounters",   # description
            default_value = 4
        )
        
        handler.add_method_parameter(
            "get_begin_time",
            "set_begin_time",
            "begin_time",
            "model time at which to start the simulation",
            default_value = 0.0 | nbody_system.time
        )

        handler.add_method_parameter(
            "get_sync_time",
            "set_sync_time",
            "sync_time",
            "last model synchronization time",
            default_value = 0.0 | nbody_system.time
        )

        handler.add_method_parameter(
            "get_zero_step_mode",
            "set_zero_step_mode",
            "zero_step_mode",
            "force evolve_model to take zero-length steps",
            default_value = 0
        )

        handler.add_method_parameter(
            "get_force_sync",
            "set_force_sync",
            "force_sync",
            "force evolve_model to sync at specified time",
            default_value = 0
        )

        handler.add_method_parameter(
            "get_block_steps",
            "set_block_steps",
            "block_steps",
            "number of block steps",
            default_value = 0
        )

        handler.add_method_parameter(
            "get_total_steps",
            "set_total_steps",
            "total_steps",
            "total number of steps",
            default_value = 0
        )

        handler.add_method_parameter(
            "get_initial_timestep_fac",  # getter name in interface.cc
            "set_initial_timestep_fac",  # setter name in interface.cc
            "initial_timestep_fac",      # python parameter name
            "initial timestep factor",   # description
            default_value = 0.0625
        )

        handler.add_method_parameter(
            "get_initial_timestep_limit", # getter name in interface.cc
            "set_initial_timestep_limit", # setter name in interface.cc
            "initial_timestep_limit",     # python parameter name
            "initial timestep limit",     # description
            default_value = 0.03125
        )

        handler.add_method_parameter(
            "get_initial_timestep_median", # getter name in interface.cc
            "set_initial_timestep_median", # setter name in interface.cc
            "initial_timestep_median",     # python parameter name
            "initial timestep median factor", # description
            default_value = 8.0
        )

        self.stopping_conditions.define_parameters(handler)
        
    def update_particle_set(self):
        """
        update the particle set after changes in the code
        
        this implementation needs to move to the
        amuse.datamodel.incode_storage module, as
        it uses a lot of internal methods and info!
        
        """
        number_of_updated_particles = self.get_number_of_particles_updated()
        
        if number_of_updated_particles == 0:
            return
        
        indices_in_update_list = list(range(number_of_updated_particles))
        particle_indices, updates \
            = self.get_id_of_updated_particle(indices_in_update_list)
        
        incode_storage = self.particles._private.attribute_storage
        
        indices_to_remove = []
        indices_to_add = []
        for index, status in zip(particle_indices, updates):
            if status == 1:                        # deletion
                indices_to_remove.append(index)
            elif status == 2:                        # addition
                indices_to_add.append(index)

        print('')
        print("indices_to_remove:", indices_to_remove)
        print("indices_to_add:", indices_to_add)

        if len(indices_to_remove) > 0:
            incode_storage._remove_indices(indices_to_remove)
        if len(indices_to_add) > 0:
            incode_storage._add_indices(indices_to_add)
        
    def define_methods(self, handler):
        GravitationalDynamics.define_methods(self, handler)

        # Turn interface functions into methods.
        handler.add_method(
            "new_particle",
            (
                nbody_system.mass,
                nbody_system.length,
                nbody_system.length,
                nbody_system.length,
                nbody_system.speed,
                nbody_system.speed,
                nbody_system.speed,
                nbody_system.length,
                handler.NO_UNIT
            ),
            (
                handler.INDEX,
                handler.ERROR_CODE
            )
        )

        handler.add_method(
            "set_eps2",
            (
                nbody_system.length * nbody_system.length
            ),
            (
                handler.ERROR_CODE
            )
        )

        handler.add_method(
            "get_eps2",
            (),
            (
                nbody_system.length * nbody_system.length,
                handler.ERROR_CODE
            )
        )

        self.stopping_conditions.define_methods(handler)

    def define_particle_sets(self, handler):
        GravitationalDynamics.define_particle_sets(self, handler)
        
        handler.add_getter('particles', 'get_particle_timestep',
                          names = ('timestep',))
        handler.add_getter('particles', 'get_potential',
                          names=('potential_in_code',))
        
        self.stopping_conditions.define_particle_set(handler)


Ph4 = ph4



#!/usr/bin/env python

import getopt
import string
from pylab import *
import re
import numpy
from math import log10

# Display a time sequence from a (not yet standard) AMUSE log file.

if __name__ == '__main__':
    
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], "f:lq:t:")
    except getopt.GetoptError as err:
        print(str(err))
        sys.exit(1)
    
    infile = ''
    quantity = ''
    log = 0
    lt = 'b-'
    
    for o, a in opts:
        if o == "-f":
            infile = a
        elif o == "-l":
            log = 1
        elif o == "-q":
            quantity = a
        elif o == "-t":
            lt = a
        else:
            print("unexpected argument", o)
    
    #if infile == '' or quantity == '':
        #print 'plotq -f file -q quantity -t linetype'
        #sys.exit(0)
    
    tlist = []
    qlist = []
    nq = 1

    f = open(infile, 'r')
    for line in f:
        if len(line) > 0:
            cols = line.split()
            if len(cols) > 0 and cols[0] == '%%%':
                if cols[1] == 'time=':
                    tlist.append(float(cols[2]))
                elif cols[1] == 'Multiple':
                    bin_id = int(cols[2])
                    #if log: q = log10(q)
                    strsplit = cols[10].split('=')
                    if len(strsplit) == 2:
                        mykT = float(strsplit[1])
                        mydata = [tlist[len(tlist)-1], bin_id, mykT]
                        qlist.append(mydata)
                #elif re.search(quantity+'\[.*\]'+'=', cols[1]):
                    #s = re.search(quantity+'\[.*\]'+'=', cols[1]).group()
                    # Messy! There must surely be a better way to do this...
                    #i1 = string.index(s,'[')
                    #i2 = string.index(s,']')
                    #nq = int(s[i1+1:i2])
                    #clist = []
                    #for i in range(2,2+nq):
                        #q = float(cols[i])
                        #if log: q = log10(q)
                        #clist.append(q)
                    #qlist.append(clist)
    f.close()
    
    multiple_index = []
    i=0
    while i<len(qlist):
      j=0
      original = True
      while j<len(multiple_index):
        if multiple_index[j] == qlist[i][1]:
          original = False 
        j += 1
      if original:
        multiple_index.append(qlist[i][1])
      i += 1
    
    counter = 0
    i=0
    while i<len(multiple_index):
      myidata = []
      myxdata = []
      myydata = []
      j=0
      while j<len(qlist):
        if multiple_index[i] == qlist[j][1]:
          myidata.append(multiple_index[i])
          myxdata.append(qlist[j][0])
          myydata.append(qlist[j][2])
          if qlist[j][0] == 300.0:
            counter += 1
        j += 1
      plot(myxdata, myydata)
      hold('on')
      i += 1
    
    show()
    
    
    i=0
    while i<len(myxdata):
      print(myxdata[i])
    
      i += 1
    print(counter)
    
    """
    if len(qlist) > 0:
        if log: quantity = 'log10 '+quantity
        if nq == 1:
            plot(tlist, qlist, lt)
        else:
            qplot = numpy.array(qlist)
            for i in range(nq):
                plot(tlist, qplot[:,i], lt, linewidth=1.0)
        xlabel('time')
        ylabel(quantity)
        show()
    else:
        print 'No data to plot.'
    """
